#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
from typing import List

from fabric_cf.actor.core.util.id import ID
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_kernel import OrchestratorKernelSingleton
from fabric_cf.orchestrator.core.i_status_update_callback import IStatusUpdateCallback
from fabric_cf.orchestrator.core.modify_operation import ModifyOperation
from fabric_cf.orchestrator.core.reservation_id_with_modify_index import ReservationIDWithModifyIndex


class ModifyQueueCallback(IStatusUpdateCallback):
    def __init__(self):
        self.lock = threading.Lock()
        self.modify_queue = {}

    def success(self, *, ok: List[ReservationIDWithModifyIndex], act_on: List[ID]):
        """
        Success callback
        :param ok:
        :param act_on:
        :return:
        """
        self.check_modify_queue(ok_or_failed=next(iter(ok)))

    def failure(self, *, failed: List[ReservationIDWithModifyIndex], ok: List[ReservationIDWithModifyIndex],
                act_on: List[ID]):
        """
        Failure callback
        :param failed:
        :param ok:
        :param act_on:
        :return:
        """
        self.check_modify_queue(ok_or_failed=next(iter(failed)))

    def check_modify_queue(self, *, ok_or_failed: ReservationIDWithModifyIndex):
        """
        Check Modify Queue
        :param ok_or_failed:
        :return:
        """
        try:
            self.lock.acquire()
            res_queue = self.modify_queue.get(ok_or_failed.get_reservation_id(), None)
            if res_queue is None:
                raise OrchestratorException("no queue found for {}, skipping processing".format(ok_or_failed))

            mop = res_queue.pop(0)

            if mop is None:
                raise OrchestratorException("no modify operation found at top of the queue, proceeding")

            if mop.get() != ok_or_failed:
                raise OrchestratorException("dequeued reservation {} which doesn't match expected {}".
                                            format(mop.get(),ok_or_failed))

            if len(res_queue) > 0:
                mop = res_queue.pop(0)
            else:
                mop = None

            if mop is not None:
                mop_index = self.modify_sliver(rid=ok_or_failed.get_reservation_id(),
                                               modify_sub_command=mop.get_sub_command(),
                                               properties=mop.get_properties())
                mop.override_index(mop_index)
                watch_list = [mop.get()]
                OrchestratorKernelSingleton.get().get_sut().add_modify_status_watch(watch=watch_list, act=None,
                                                                                    callback=self)
            else:
                self.modify_queue.pop(ok_or_failed.get_reservation_id())
        finally:
            self.lock.release()

    def enqueue_modify(self, *, res: str, modify_sub_command: str, properties: dict):
        """
        Enqueue Modify
        :param res:
        :param modify_sub_command:
        :param properties:
        :return:
        """
        rid = ID(uid=res)
        try:
            self.lock.acquire()
            reservation_queue = None
            if rid not in self.modify_queue:
                reservation_queue = []
            else:
                reservation_queue = self.modify_queue.get(rid)

            mop = ModifyOperation(rid=rid, index=0, modify_sub_command=modify_sub_command, properties=properties)
            reservation_queue.append(mop)

            self.modify_queue[rid] = reservation_queue

            if len(reservation_queue) == 1:
                mod_index = self.modify_sliver(rid=rid, modify_sub_command=modify_sub_command, properties=properties)
                mop.override_index(index=mod_index)
                watch_list = [mop.get()]
                OrchestratorKernelSingleton.get().get_sut().add_modify_status_watch(watch=watch_list, act=None,
                                                                                    callback=self)
        finally:
            self.lock.release()

    def modify_sliver(self, *, rid: ID, modify_sub_command: str, properties: dict) -> int:
        """
        Modify sliver
        :param rid:
        :param modify_sub_command:
        :param properties:
        :return:
        """
        try:
            controller = OrchestratorKernelSingleton.get().get_management_actor()
            reservation = controller.get_reservation(rid=rid)
            if reservation is None:
                raise OrchestratorException("Unable to find reservation {}".format(rid))

            config_props = reservation.get_config_properties()
            if config_props is None:
                raise OrchestratorException("Unable to get configuration properties for reservation {}".format(rid))

            # Update the properties
            # TODO
            index = 0

            controller.modify_reservation(rid=rid, modify_properties=properties)
            return index
        except Exception as e:
            raise OrchestratorException("Unable to modify sliver reservation: {}".format(e))


class ModifyQueueCallbackSingleton:
    __instance = None

    def __init__(self):
        if self.__instance is not None:
            raise OrchestratorException("Singleton can't be created twice !")

    def get(self):
        """
        Actually create an instance
        """
        if self.__instance is None:
            self.__instance = ModifyQueueCallback()
        return self.__instance

    get = classmethod(get)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING

from fabric_mb.message_bus.messages.result_reservation_avro import ResultReservationAvro
from fabric_mb.message_bus.messages.result_avro import ResultAvro
from fabric_mb.message_bus.messages.result_units_avro import ResultUnitsAvro

from fabric_cf.actor.core.apis.abc_authority import ABCAuthority
from fabric_cf.actor.core.common.constants import Constants, ErrorCodes
from fabric_cf.actor.core.common.exceptions import ReservationNotFoundException
from fabric_cf.actor.core.manage.converter import Converter
from fabric_cf.actor.core.manage.management_object import ManagementObject
from fabric_cf.actor.core.manage.proxy_protocol_descriptor import ProxyProtocolDescriptor
from fabric_cf.actor.core.manage.server_actor_management_object import ServerActorManagementObject
from fabric_cf.actor.security.access_checker import AccessChecker
from fabric_cf.actor.security.pdp_auth import ActionId, ResourceType

if TYPE_CHECKING:
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.apis.abc_substrate_database import ABCSubstrateDatabase
    from fabric_cf.actor.core.util.id import ID


class AuthorityManagementObject(ServerActorManagementObject):
    def __init__(self, *, authority: ABCAuthority = None):
        super().__init__(sa=authority)

    def register_protocols(self):
        from fabric_cf.actor.core.manage.local.local_authority import LocalAuthority
        local = ProxyProtocolDescriptor(protocol=Constants.PROTOCOL_LOCAL,
                                        proxy_class=LocalAuthority.__name__,
                                        proxy_module=LocalAuthority.__module__)

        from fabric_cf.actor.core.manage.kafka.kafka_authority import KafkaAuthority
        kakfa = ProxyProtocolDescriptor(protocol=Constants.PROTOCOL_KAFKA, proxy_class=KafkaAuthority.__name__,
                                        proxy_module=KafkaAuthority.__module__)

        self.proxies = []
        self.proxies.append(local)
        self.proxies.append(kakfa)

    def save(self) -> dict:
        properties = super().save()
        properties[Constants.PROPERTY_CLASS_NAME] = AuthorityManagementObject.__name__
        properties[Constants.PROPERTY_MODULE_NAME] = AuthorityManagementObject.__module__

        return properties

    def get_authority_reservations(self, *, caller: AuthToken, id_token: str = None) -> ResultReservationAvro:
        result = ResultReservationAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result
        try:
            AccessChecker.check_access(action_id=ActionId.query, resource_type=ResourceType.sliver,
                                       token=id_token, logger=self.logger, actor_type=self.actor.get_type())
            res_list = None
            try:
                res_list = self.db.get_authority_reservations()
            except Exception as e:
                self.logger.error("get_authority_reservations:db access {}".format(e))
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.interpret(exception=e))
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)
                return result

            if res_list is not None:
                result.reservations = []
                for r in res_list:
                    slice_obj = self.get_slice_by_guid(guid=r.get_slice_id())
                    r.restore(actor=self.actor, slice_obj=slice_obj, logger=self.logger)
                    rr = Converter.fill_reservation(reservation=r, full=True)
                    result.reservations.append(rr)
        except ReservationNotFoundException as e:
            self.logger.error("get_authority_reservations: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorNoSuchReservation.value)
            result.status.set_message(e.text)
        except Exception as e:
            self.logger.error("get_authority_reservations: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_substrate_database(self) -> ABCSubstrateDatabase:
        return self.actor.get_plugin().get_database()

    def get_reservation_units(self, *, caller: AuthToken, rid: ID, id_token: str = None) -> ResultUnitsAvro:
        result = ResultUnitsAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result
        try:
            AccessChecker.check_access(action_id=ActionId.query, resource_type=ResourceType.sliver,
                                       token=id_token, logger=self.logger, actor_type=self.actor.get_type())
            units_list = None
            try:
                units_list = self.db.get_units(rid=rid)
            except Exception as e:
                self.logger.error("get_reservation_units:db access {}".format(e))
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.interpret(exception=e))
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)
                return result

            if units_list is not None:
                result.units = Converter.fill_units(unit_list=units_list)
        except Exception as e:
            self.logger.error("get_reservation_units: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_reservation_unit(self, *, caller: AuthToken, uid: ID, id_token: str = None) -> ResultUnitsAvro:
        result = ResultUnitsAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            return result
        try:
            AccessChecker.check_access(action_id=ActionId.query, resource_type=ResourceType.sliver,
                                       token=id_token, logger=self.logger, actor_type=self.actor.get_type())
            units_list = None
            try:
                unit = self.db.get_unit(uid=uid)
                units_list = [unit]
            except Exception as e:
                self.logger.error("get_reservation_unit:db access {}".format(e))
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.interpret(exception=e))
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)
                return result

            if units_list is not None:
                result.units = Converter.fill_units(unit_list=units_list)
        except Exception as e:
            self.logger.error("get_reservation_unit: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.interpret())
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import queue
import threading
import traceback
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_actor_mixin import ActorType
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.exceptions import ControllerException
from fabric_cf.actor.core.manage.controller_management_object import ControllerManagementObject
from fabric_cf.actor.core.manage.kafka.services.kafka_controller_service import KafkaControllerService
from fabric_cf.actor.core.proxies.kafka.services.controller_service import ControllerService
from fabric_cf.actor.core.apis.abc_controller import ABCController
from fabric_cf.actor.core.core.actor import ActorMixin
from fabric_cf.actor.core.registry.peer_registry import PeerRegistry
from fabric_cf.actor.core.util.reservation_set import ReservationSet
from fabric_cf.actor.core.apis.abc_controller_reservation import ABCControllerReservation

if TYPE_CHECKING:
    from fabric_cf.actor.core.time.actor_clock import ActorClock
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.apis.abc_broker_proxy import ABCBrokerProxy
    from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice
    from fabric_cf.actor.core.util.id import ID


class Controller(ActorMixin, ABCController):
    """
    Implements Controller
    """
    saved_extended_renewable = ReservationSet()

    def __init__(self, *, identity: AuthToken = None, clock: ActorClock = None):
        super().__init__(auth=identity, clock=clock)
        # Recovered reservations that need to obtain tickets.
        self.ticketing = ReservationSet()
        # Recovered reservations that need to extend tickets.
        self.extending_ticket = ReservationSet()
        # Recovered reservations that need to be redeemed.
        self.redeeming = ReservationSet()
        # Recovered reservations that need to extend leases.
        self.extending_lease = ReservationSet()
        # Recovered reservations that need to modify leases
        self.modifying_lease = ReservationSet()
        # Peer registry.
        self.registry = PeerRegistry()
        # initialization status
        self.initialized = False
        self.type = ActorType.Orchestrator

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['recovered']
        del state['wrapper']
        del state['logger']
        del state['clock']
        del state['current_cycle']
        del state['first_tick']
        del state['stopped']
        del state['initialized']
        del state['thread_lock']
        del state['thread']
        del state['timer_queue']
        del state['event_queue']
        del state['reservation_tracker']
        del state['subscription_id']
        del state['actor_main_lock']
        del state['closing']
        del state['message_service']

        del state['ticketing']
        del state['extending_ticket']
        del state['redeeming']
        del state['extending_lease']
        del state['modifying_lease']
        del state['registry']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.recovered = False
        self.wrapper = None
        self.logger = None
        self.clock = None
        self.current_cycle = -1
        self.first_tick = True
        self.stopped = False
        self.initialized = False
        self.thread = None
        self.thread_lock = threading.Lock()
        self.timer_queue = queue.Queue()
        self.event_queue = queue.Queue()
        self.reservation_tracker = None
        self.subscription_id = None
        self.actor_main_lock = threading.Condition()
        self.closing = ReservationSet()
        self.message_service = None

        self.ticketing = ReservationSet()
        self.extending_ticket = ReservationSet()
        self.redeeming = ReservationSet()
        self.extending_lease = ReservationSet()
        self.modifying_lease = ReservationSet()
        self.registry = PeerRegistry()

    def actor_added(self):
        super().actor_added()
        self.registry.actor_added()

    def add_broker(self, *, broker: ABCBrokerProxy):
        self.registry.add_broker(broker=broker)

    def bid(self):
        """
        Bids for resources as dictated by the plugin bidding policy for the
        current cycle.

        @throws Exception in case of error
        """
        # Invoke policy module to select candidates for ticket and extend. Note
        # that candidates structure is discarded when we're done.
        candidates = self.policy.formulate_bids(cycle=self.current_cycle)
        if candidates is not None:
            # Issue new ticket requests.
            ticketing = candidates.get_ticketing()
            if ticketing is not None:
                for ticket in ticketing.values():
                    try:
                        self.wrapper.ticket(reservation=ticket, destination=self)
                    except Exception as e:
                        self.logger.error("unexpected ticket failure for #{} {}".format(ticket.get_reservation_id(), e))
                        ticket.fail(message="unexpected ticket failure {}".format(e))

            extending = candidates.get_extending()
            if extending is not None:
                for extend in extending.values():
                    try:
                        self.wrapper.extend_ticket(reservation=extend)
                    except Exception as e:
                        self.logger.error("unexpected extend failure for #{} {}".format(extend.get_reservation_id(), e))
                        extend.fail(message="unexpected extend failure {}".format(e))

    def claim_delegation_client(self, *, delegation_id: str = None, slice_object: ABCSlice = None,
                                broker: ABCBrokerProxy = None, id_token: str = None) -> ABCDelegation:
        raise ControllerException("Not implemented")

    def reclaim_delegation_client(self, *, delegation_id: str = None, slice_object: ABCSlice = None,
                                  broker: ABCBrokerProxy = None, id_token: str = None) -> ABCDelegation:
        raise ControllerException("Not implemented")

    def close_expiring(self):
        """
        Issues close requests on all reservations scheduled for closing on the
        current cycle
        """
        rset = self.policy.get_closing(cycle=self.current_cycle)

        if rset is not None and rset.size() > 0:
            self.logger.info("SlottedSM close expiring for cycle {} expiring {}".format(self.current_cycle, rset))
            self.close_reservations(reservations=rset)

    def demand(self, *, rid: ID):
        if rid is None:
            raise ControllerException("Invalid argument")

        reservation = self.get_reservation(rid=rid)

        if reservation is None:
            raise ControllerException("Unknown reservation {}".format(rid))

        self.policy.demand(reservation=reservation)
        reservation.set_policy(policy=self.policy)

    def extend_lease_reservation(self, *, reservation: ABCControllerReservation):
        """
        Extend Lease for a reservation
        @param reservation reservation
        """
        if not self.recovered:
            self.extending_lease.add(reservation=reservation)
        else:
            self.wrapper.extend_lease(reservation=reservation)

    def extend_lease(self, *, reservation: ABCControllerReservation = None, rset: ReservationSet = None):
        if reservation is not None and rset is not None:
            raise ControllerException("Invalid Arguments: reservation and rset can not be both not None")
        if reservation is None and rset is None:
            raise ControllerException("Invalid Arguments: reservation and rset can not be both None")

        if reservation is not None:
            self.extend_lease_reservation(reservation=reservation)

        if rset is not None:
            for r in rset.values():
                try:
                    if isinstance(r, ABCControllerReservation):
                        self.extend_lease_reservation(reservation=r)
                    else:
                        self.logger.warning("Reservation #{} cannot extendLease".format(
                            r.get_reservation_id()))
                except Exception as e:
                    self.logger.error("Could not extend_lease for #{} e={}".format(
                        r.get_reservation_id(), e))

    def extend_ticket_client(self, *, reservation: ABCClientReservation):
        if not self.recovered:
            self.extending_ticket.add(reservation=reservation)
        else:
            self.wrapper.extend_ticket(reservation=reservation)

    def extend_tickets_client(self, *, rset: ReservationSet):
        for reservation in rset.values():
            try:
                if isinstance(reservation, ABCClientReservation):
                    self.extend_ticket_client(reservation=reservation)
                else:
                    self.logger.warning("Reservation # {} cannot be ticketed".format(reservation.get_reservation_id()))
            except Exception as e:
                self.logger.error("Could not ticket for #{} e: {}".format(reservation.get_reservation_id(), e))

    def get_broker(self, *, guid: ID) -> ABCBrokerProxy:
        return self.registry.get_broker(guid=guid)

    def get_brokers(self) -> list:
        return self.registry.get_brokers()

    def get_default_broker(self) -> ABCBrokerProxy:
        return self.registry.get_default_broker()

    def initialize(self):
        if not self.initialized:
            super().initialize()

            self.registry.set_slices_plugin(plugin=self.plugin)
            self.registry.initialize()

            self.initialized = True

    def process_redeeming(self):
        """
        Issue redeem requests on all reservations scheduled for redeeming on the current cycle
        """
        rset = self.policy.get_redeeming(cycle=self.current_cycle)

        if rset is not None and rset.size() > 0:
            self.logger.info("SlottedController redeem for cycle {} redeeming {}".format(self.current_cycle, rset))

            self.redeem_reservations(rset=rset)

    def redeem(self, *, reservation: ABCControllerReservation):
        if not self.recovered:
            self.redeeming.add(reservation=reservation)
        else:
            self.wrapper.redeem(reservation=reservation)

    def redeem_reservations(self, *, rset: ReservationSet):
        for reservation in rset.values():
            try:
                if isinstance(reservation, ABCControllerReservation):
                    self.redeem(reservation=reservation)
                else:
                    self.logger.warning("Reservation #{} cannot be redeemed".format(reservation.get_reservation_id()))
            except Exception as e:
                self.logger.error(traceback.format_exc())
                self.logger.error("Could not redeem for #{} {}".format(reservation.get_reservation_id(), e))

    def ticket_client(self, *, reservation: ABCClientReservation):
        if not self.recovered:
            self.ticketing.add(reservation=reservation)
        else:
            self.wrapper.ticket(reservation=reservation, destination=self)

    def tickets_client(self, *, rset: ReservationSet):
        for reservation in rset.values():
            try:
                if isinstance(reservation, ABCClientReservation):
                    self.ticket_client(reservation=reservation)
                else:
                    self.logger.warning("Reservation #{} cannot be ticketed".format(reservation.get_reservation_id()))
            except Exception as e:
                self.logger.error("Could not ticket for #{} e: {}".format(reservation.get_reservation_id(), e))

    def tick_handler(self):
        self.close_expiring()
        self.process_redeeming()
        self.bid()

    def update_lease(self, *, reservation: ABCReservationMixin, update_data, caller: AuthToken):
        if not self.is_recovered() or self.is_stopped():
            raise ControllerException("This actor cannot receive calls")

        self.wrapper.update_lease(reservation=reservation, update_data=update_data, caller=caller)

    def update_ticket(self, *, reservation: ABCReservationMixin, update_data, caller: AuthToken):
        if not self.is_recovered() or self.is_stopped():
            raise ControllerException("This actor cannot receive calls")

        self.wrapper.update_ticket(reservation=reservation, update_data=update_data, caller=caller)

    def update_delegation(self, *, delegation: ABCDelegation, update_data, caller: AuthToken):
        raise ControllerException("Not supported in controller")

    def modify(self, *, reservation_id: ID, modify_properties: dict):
        if reservation_id is None or modify_properties is None:
            self.logger.error("modifyProperties argument is null or non-existing reservation")

        rc = None
        try:
            rc = self.get_reservation(rid=reservation_id)
        except Exception as e:
            self.logger.error("Could not find reservation #{} e: {}".format(reservation_id, e))

        if rc is None:
            raise ControllerException("Unknown reservation: {}".format(reservation_id))

        if rc.get_resources() is not None:
            # TODO
            print("TODO")
        else:
            self.logger.warning("There are no approved resources for {}, no modify properties will be added".format(
                reservation_id))

        if not self.recovered:
            self.modifying_lease.add(reservation=rc)
        else:
            self.wrapper.modify_lease(reservation=rc)

    def save_extending_renewable(self):
        """
        For recovery, mark extending reservations renewable or the opposite
        and save this, then restore afterwards
        """
        for reservation in self.extending_ticket.values():
            try:
                if isinstance(reservation, ABCClientReservation):
                    if not reservation.get_renewable():
                        reservation.set_renewable(renewable=True)
                        self.saved_extended_renewable.add(reservation=reservation)
                else:
                    self.logger.warning("Reservation #{} cannot be remarked".format(reservation.get_reservation_id()))
            except Exception as e:
                self.logger.error("Could not mark ticket renewable for #{} e: {}".format(
                    reservation.get_reservation_id(), e))

    def restore_extending_renewable(self):
        """
        Restore the value of renewable field after recovery if we changed it
        """
        for reservation in self.saved_extended_renewable.values():
            try:
                reservation.set_renewable(renewable=False)
            except Exception as e:
                self.logger.error("Could not remark ticket non renewable for #{} e: {}".format(
                    reservation.get_reservation_id(), e))

    def issue_delayed(self):
        super().issue_delayed()
        self.tickets_client(rset=self.ticketing)
        self.ticketing.clear()

        self.save_extending_renewable()
        self.extend_tickets_client(rset=self.extending_ticket)
        self.extending_ticket.clear()

        self.redeem_reservations(rset=self.redeeming)
        self.redeeming.clear()

        self.extend_lease(rset=self.extending_lease)
        self.extending_lease.clear()

    @staticmethod
    def get_management_object_class() -> str:
        return ControllerManagementObject.__name__

    @staticmethod
    def get_management_object_module() -> str:
        return ControllerManagementObject.__module__

    @staticmethod
    def get_kafka_service_class() -> str:
        return ControllerService.__name__

    @staticmethod
    def get_kafka_service_module() -> str:
        return ControllerService.__module__

    @staticmethod
    def get_mgmt_kafka_service_class() -> str:
        return KafkaControllerService.__name__

    @staticmethod
    def get_mgmt_kafka_service_module() -> str:
        return KafkaControllerService.__module__



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_policy import ABCPolicy

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_broker_reservation import ABCBrokerReservation
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin


class ABCServerPolicy(ABCPolicy):
    """
    IServerPolicy defines the policy interface for an actor acting
    as a server for another actor (broker or a site authority).
    """

    @abstractmethod
    def allocate(self, *, cycle: int):
        """
        Allocates resources to all clients who have requested them. This method
        is called by the policy once per cycle. The method should determine
        whether to perform resource allocation on the given cycle and what
        requests to consider in that process.

        @param cycle : the cycle for this allocation

        @raises Exception in case of error
        """

    @abstractmethod
    def bind(self, *, reservation: ABCReservationMixin) -> bool:
        """
        Handles an incoming request to allocate resources and issue a ticket for
        the reservation. The requested resources can be obtained by calling
        reservation.get_requested_resources(). The requested lease
        term can be obtained by calling
        reservation.get_requested_term(). Properties specific to the
        allocation protocol can be obtained by calling
        reservation.get_requested_resources().get_request_properties()
        If the policy completed processing this request, the functions should
        return true. If no further intervention is required, e.g., approval by an
        administrator, the policy should also clear the bid_pending flag.

        The policy may decide to defer the request for a later time. In this case
        the function should return false and the bid_pending flag
        should remain unchanged.

        This method may be invoked multiple times for a given reservation, i.e.,
        if the policy delays the allocation, the system will continue invoking
        this method at later times until the policy completes processing this
        request.

        @param reservation: reservation to allocate resources for.

        @return true, if the request has been fulfilled, false, if the allocation
                of resources will be delayed until a later time.
        @raises Exception in case of error
        """

    @abstractmethod
    def bind_delegation(self, *, delegation: ABCDelegation) -> bool:
        """
        Handles an incoming request to allocate resources and issue a ticket for the delegation.
        @param delegation: delegation to allocate resources for.

        @return true, if the request has been fulfilled, false, if the allocation
                of resources will be delayed until a later time.
        @raises Exception in case of error
        """

    @abstractmethod
    def donate_delegation(self, *, delegation: ABCDelegation):
        """
        Accepts ticketed resources to be used for allocation of client requests.
        The policy should add the resources represented by this delegation to
        its inventory.

        @param delegation : delegation representing resources to be used for allocation

        @raises Exception in case of error
        """

    @abstractmethod
    def extend_broker(self, *, reservation: ABCBrokerReservation) -> bool:
        """
        Handles an incoming request to extend previously allocated resources and
        issue a ticket for the reservation. The requested resources can be
        obtained by calling reservation.getRequestedResources()
        Properties specific to the allocation protocol can be obtained by calling
        reservation.get_requested_resources().get_request_properties().
        The requested lease term can be obtained by calling
        reservation.get_requested_term(). The new term must extend the
        currently allocated term.

        If the policy completed processing this request, the functions should
        return true. If no further intervention is required, e.g., approval by an
        administrator, the policy should also clear the bid_pending
        flag.

        The policy may decide to defer the request for a later time. In this case
        the function should return false and the bid_pending flag
        should remain unchanged.

        This method may be invoked multiple times for a given reservation, i.e.,
        if the policy delays the allocation, the system will continue invoking
        this method at later times until the policy completes processing this
        request.

        While the policy is free to modify the term as it wishes, care must be
        taken that the client reservation is not closed before the extension is
        actually granted.

        @param reservation : reservation to allocate resources for.

        @return true, if the request has been fulfilled, false, if the allocation
                of resources will be delayed until a later time.
        @raises Exception in case of error
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
import traceback
from typing import List

from fabric_cf.actor.core.apis.abc_base_plugin import ABCBasePlugin
from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import ReservationNotFoundException, DelegationNotFoundException, \
    KernelException
from fabric_cf.actor.core.kernel.authority_reservation import AuthorityReservation
from fabric_cf.actor.core.kernel.failed_rpc import FailedRPC
from fabric_cf.actor.core.apis.abc_kernel_controller_reservation_mixin import ABCKernelControllerReservationMixin
from fabric_cf.actor.core.apis.abc_kernel_reservation import ABCKernelReservation
from fabric_cf.actor.core.apis.abc_kernel_server_reservation import ABCKernelServerReservationMixin
from fabric_cf.actor.core.apis.abc_kernel_slice import ABCKernelSlice
from fabric_cf.actor.core.kernel.request_types import RequestTypes
from fabric_cf.actor.core.kernel.reservation import Reservation
from fabric_cf.actor.core.kernel.reservation_states import ReservationPendingStates, ReservationStates
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.kernel.sequence_comparison_codes import SequenceComparisonCodes
from fabric_cf.actor.core.kernel.slice_state_machine import SliceStateMachine
from fabric_cf.actor.core.kernel.slice_table import SliceTable
from fabric_cf.actor.core.time.term import Term
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.reservation_set import ReservationSet
from fabric_cf.actor.core.util.update_data import UpdateData
from fabric_cf.actor.security.auth_token import AuthToken


class Kernel:
    def __init__(self, *, plugin: ABCBasePlugin, policy: ABCPolicy, logger):
        # The plugin.
        self.plugin = plugin
        # Policy
        self.policy = policy
        # Logger
        self.logger = logger
        # All slices managed by the kernel
        self.slices = SliceTable()
        # All reservations managed by the kernel.
        self.reservations = ReservationSet()
        self.delegations = {}
        self.lock = threading.Lock()
        self.nothing_pending = threading.Condition()

    def amend_reserve(self, *, reservation: ABCKernelReservation):
        """
        Amends a previous reserve operation (both client and server side) for the
        reservation.
        @param reservation reservation
        @throws Exception
        """
        try:
            reservation.reserve(policy=self.policy)
            self.plugin.get_database().update_reservation(reservation=reservation)
            if not reservation.is_failed():
                reservation.service_reserve()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            err = f"An error occurred during amend reserve for reservation #{reservation.get_reservation_id()}"
            self.error(err=err, e=e)

    def amend_delegate(self, *, delegation: ABCDelegation):
        """
        Amends a previous delegate operation for the delegation.
        @param delegation delegation
        @throws Exception
        """
        try:
            delegation.delegate(policy=self.policy)
            self.plugin.get_database().update_delegation(delegation=delegation)
            if not delegation.is_closed():
                delegation.service_delegate()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            err = f"An error occurred during amend delegate for delegation #{delegation.get_delegation_id()}"
            self.error(err=err, e=e)

    def claim_delegation(self, *, delegation: ABCDelegation):
        """
        Processes a requests to claim new ticket for previously exported
        resources (broker role). On the client side this request is issued by
        @param delegation the delegation being claimed
        @throws Exception
        """
        try:
            delegation.claim()
            self.plugin.get_database().update_delegation(delegation=delegation)
        except Exception as e:
            err = f"An error occurred during claim for delegation #{delegation.get_delegation_id()}"
            self.logger.error(traceback.format_exc())
            self.error(err=err, e=e)

    def reclaim_delegation(self, *, delegation: ABCDelegation, id_token: str):
        """
        Processes a requests to claim new ticket for previously exported
        resources (broker role). On the client side this request is issued by
        @param delegation the delegation being claimed
        @param id_token id token
        @throws Exception
        """
        try:
            delegation.reclaim(id_token=id_token)
            self.plugin.get_database().update_delegation(delegation=delegation)
        except Exception as e:
            err = f"An error occurred during reclaim for delegation #{delegation.get_delegation_id()}"
            self.logger.error(traceback.format_exc())
            self.error(err=err, e=e)

    def fail(self, *, reservation: ABCKernelReservation, message: str):
        """
        Handle a failed reservation
        @param reservation reservation
        @param message message
        """
        if not reservation.is_failed() and not reservation.is_closed():
            reservation.fail(message=message, exception=None)
        self.plugin.get_database().update_reservation(reservation=reservation)

    def fail_delegation(self, *, delegation: ABCDelegation, message: str):
        """
        Handle a failed delegation
        @param delegation delegation
        @param message message
        """
        if not delegation.is_failed() and not delegation.is_closed():
            delegation.fail(message=message, exception=None)
        self.plugin.get_database().update_delegation(delegation=delegation)

    def close(self, *, reservation: ABCKernelReservation):
        """
        Handles a close operation for the reservation.
        Client: perform local close operations and issue close request to
        authority.
        Broker: perform local close operations
        Authority: process a close request
        @param reservation reservation for which to perform close
        @throws Exception
        """
        try:
            if not reservation.is_closed() and not reservation.is_closing():
                self.policy.close(reservation=reservation)
                reservation.close()
                self.plugin.get_database().update_reservation(reservation=reservation)
                reservation.service_close()
        except Exception as e:
            err = f"An error occurred during close for reservation #{reservation.get_reservation_id()}"
            self.logger.error(traceback.format_exc())
            self.error(err=err, e=e)

    def compare_and_update(self, *, incoming: ABCKernelServerReservationMixin,
                           current: ABCKernelServerReservationMixin):
        """
        Compares the incoming request to the corresponding reservation stored at
        this actor. First compares sequence numbers. If the incoming request has
        a larger sequence number and there is no pending operation for this
        reservation, we update the sequence number of the current reservation and
        set the requestedTerm and requestedResources fields.

        @param incoming the incoming request
        @param current the corresponding reservation stored at the server
        @return a comparison status flag (see Sequence*)
        """
        code = SequenceComparisonCodes.SequenceEqual
        if current.get_sequence_in() < incoming.get_sequence_in():
            if current.is_no_pending():
                code = SequenceComparisonCodes.SequenceGreater
                current.set_sequence_in(sequence=incoming.get_sequence_in())
                current.set_requested_resources(resources=incoming.get_requested_resources())
                current.set_requested_term(term=incoming.get_requested_term())
        else:
            if current.get_sequence_in() > incoming.get_sequence_in():
                code = SequenceComparisonCodes.SequenceSmaller
        return code

    @staticmethod
    def compare_and_update_ignore_pending(*, incoming: ABCKernelServerReservationMixin,
                                          current: ABCKernelServerReservationMixin):
        """
        Compares the incoming request to the corresponding reservation stored at
        this actor. First compares sequence numbers. If the incoming request has
        a larger sequence number and there is no pending operation for this
        reservation, we update the sequence number of the current reservation and
        set the requestedTerm and requestedResources fields.

        @param incoming the incoming request
        @param current the corresponding reservation stored at the server
        @return a comparison status flag (see Sequence*)
        """
        code = SequenceComparisonCodes.SequenceEqual
        if current.get_sequence_in() < incoming.get_sequence_in():
            code = SequenceComparisonCodes.SequenceGreater
            current.set_sequence_in(sequence=incoming.get_sequence_in())
            current.set_requested_resources(resources=incoming.get_requested_resources())
            current.set_requested_term(term=incoming.get_requested_term())
        else:
            if current.get_sequence_in() > incoming.get_sequence_in():
                code = SequenceComparisonCodes.SequenceSmaller
        return code

    def error(self, *, err: str, e: Exception):
        """
        Logs the specified exception and re-throws it.
        @param err error message
        @param e exception
        """
        self.logger.error(f"Error: {err} Exception: {e}")
        raise e

    def extend_lease(self, *, reservation: ABCKernelReservation):
        """
        Handles an extend lease operation for the reservation.
        Client: issue an extend lease request.
        Authority: process a request for a lease extension.
        @param reservation reservation for which to perform extend lease
        @throws Exception
        """
        try:
            reservation.extend_lease()
            self.plugin.get_database().update_reservation(reservation=reservation)
            if not reservation.is_failed():
                reservation.service_extend_lease()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during extend lease for reservation #{reservation.get_reservation_id()}",
                       e=e)

    def modify_lease(self, *, reservation: ABCKernelReservation):
        """
        Handles a modify lease operation for the reservation.
        Client: issue a modify lease request.

        Authority: process a request for a modifying a lease.
        @param reservation reservation for which to perform extend lease
        @throws Exception
        """
        try:
            reservation.modify_lease()
            self.plugin.get_database().update_reservation(reservation=reservation)
            if not reservation.is_failed():
                reservation.service_modify_lease()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during modify lease for reservation #{reservation.get_reservation_id()}",
                       e=e)

    def extend_reservation(self, *, rid: ID, resources: ResourceSet, term: Term) -> int:
        """
        Extends the reservation with the given resources and term.
        @param rid reservation identifier of reservation to extend
        @param resources resources to use for the extension
        @param term term to use for the extension
        @return 0 if the reservation extension operation can be initiated,
                if the reservation has a pending operation, which prevents the extend
                operation from being initiated.
        @throws Exception
        """
        real = self.reservations.get(rid=rid)
        ticket = True

        if real is None:
            raise KernelException(f"Unknown reservation rid: {rid}")

        # check for a pending operation: we cannot service the extend if there is another operation in progress.
        if real.get_pending_state() != ReservationPendingStates.None_:
            return Constants.RESERVATION_HAS_PENDING_OPERATION

        # attach the desired extension term and resource set
        real.set_approved(term=term, approved_resources=resources)
        # notify the policy that a reservation is about to be extended
        self.policy.extend(reservation=real, resources=resources, term=term)

        if isinstance(real, AuthorityReservation):
            ticket = False

        if ticket:
            real.extend_ticket(actor=self.plugin.get_actor())
        else:
            real.extend_lease()

        self.plugin.get_database().update_reservation(reservation=real)

        if not real.is_failed():
            if ticket:
                real.service_extend_ticket()
            else:
                real.service_extend_lease()

        return 0

    def extend_ticket(self, *, reservation: ABCKernelReservation):
        """
        Handles an extend ticket operation for the reservation.
        Client: issue an extend ticket request.
        Broker: process a request for a ticket extension.
        @param reservation reservation for which to perform extend ticket
        @throws Exception
        """
        try:
            self.logger.debug(f"Processing extend ticket for reservation={type(reservation)}")
            if reservation.can_renew():
                reservation.extend_ticket(actor=self.plugin.get_actor())
            else:
                raise KernelException("The reservation state prevents it from extending its ticket.")

            self.plugin.get_database().update_reservation(reservation=reservation)

            if not reservation.is_failed():
                reservation.service_extend_ticket()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during extend ticket for reservation #{reservation.get_reservation_id()}",
                       e=e)

    def get_client_slices(self) -> List[ABCKernelSlice]:
        """
        Returns all client slices.
        @return an array of client slices
        """
        return self.slices.get_client_slices()

    def get_inventory_slices(self) -> List[ABCKernelSlice]:
        """
        Returns all inventory slices.
        @return an array of inventory slices
        """
        return self.slices.get_inventory_slices()

    def get_local_slice(self, *, slice_object: ABCSlice) -> ABCKernelSlice:
        """
        Returns the slice object registered with the kernel that corresponds to
        the argument.
        @param slice_object incoming slice object
        @return the locally registered slice object
        @throws IllegalArgumentException if the arguments are invalid
        @throws Exception if no locally registered slice object exists
        """
        if slice_object is None or slice_object.get_slice_id() is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        return self.slices.get(slice_id=slice_object.get_slice_id(), raise_exception=True)

    def get_or_create_local_slice(self, *, identity: AuthToken, reservation: ABCKernelReservation,
                                  create_new_slice: bool) -> ABCKernelSlice:
        """
        Returns the slice specified in the reservation or creates a new slice
        with the given parameters. Newly created slices are registered with the
        kernel.
        @param identity actor identity
        @param reservation reservation
        @param create_new_slice true if slice to be created if not found
        @return the slice object
        @throws Exception
        """
        slice_name = reservation.get_slice().get_name()
        slice_id = reservation.get_slice().get_slice_id()
        config_properties = reservation.get_slice().get_config_properties()

        result = self.get_slice(slice_id=slice_id)
        if result is None:
            if create_new_slice:
                result = self.plugin.create_slice(slice_id=slice_id, name=slice_name)
                result.set_config_properties(value=config_properties)
                if reservation.get_slice().is_broker_client():
                    result.set_broker_client()
                else:
                    if reservation.get_slice().is_client():
                        result.set_client()
            else:
                result = reservation.get_kernel_slice()

            result.set_owner(owner=identity)
            self.register_slice(slice_object=result)
        return result

    def get_reservation(self, *, rid: ID) -> ABCKernelReservation:
        """
        Returns the specified reservation.
        @param rid reservation id
        @return reservation
        """
        if rid is not None:
            return self.reservations.get(rid=rid)
        return None

    def get_reservations(self, *, slice_id: ID) -> List[ABCKernelReservation]:
        """
        Returns all reservations in the specified slice.
        @param slice_id slice id
        @return an array of reservations
        """
        result = None
        if slice_id is not None:
            sl = self.slices.get(slice_id=slice_id)
            result = sl.get_reservations_list()
        return result

    def get_delegation(self, *, did: str) -> ABCDelegation:
        """
        Returns the specified delegation.
        @param did delegation id
        @return delegation
        """
        try:
            if did is not None:
                self.lock.acquire()
                return self.delegations.get(did, None)
        finally:
            if self.lock.locked():
                self.lock.release()
        return None

    def get_plugin(self) -> ABCBasePlugin:
        """
        Returns the plugin.
        @return plugin object
        """
        return self.plugin

    def get_slice(self, *, slice_id: ID) -> ABCKernelSlice:
        """
        Returns a slice previously registered with the kernel.
        @param slice_id slice identifier
        @return slice object
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        return self.slices.get(slice_id=slice_id)

    def is_known_slice(self, *, slice_id: ID) -> bool:
        """
        Is slice known
        @param slice_id slice id
        @return true if slice exists; false otherwise
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)
        return self.slices.contains(slice_id=slice_id)

    def get_slices(self) -> list:
        """
        Returns all registered slices.
        @return an array of slices
        """
        return self.slices.get_slices()

    def handle_duplicate_request(self, *, current: ABCKernelReservation, operation: RequestTypes):
        """
        Handles a duplicate request.
        @param current reservation
        @param operation operation code
        """
        current.handle_duplicate_request(operation=operation)

    def probe_pending_slices(self, *, slice_obj: ABCKernelSlice):
        """
        Probes to check for completion of pending operation.
        @param slice_obj the slice_obj being probed
        @throws Exception rare
        """
        try:
            if slice_obj.is_broker_client():
                return
            state_changed, slice_state = slice_obj.transition_slice(operation=SliceStateMachine.REEVALUATE)
            if state_changed:
                slice_obj.set_dirty()
            self.plugin.get_database().update_slice(slice_object=slice_obj)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during probe pending for slice_obj #{slice_obj.get_slice_id()}", e=e)

    def probe_pending(self, *, reservation: ABCKernelReservation):
        """
        Probes to check for completion of pending operation.
        @param reservation the reservation being probed
        @throws Exception rare
        """
        try:
            reservation.prepare_probe()
            reservation.probe_pending()
            self.plugin.get_database().update_reservation(reservation=reservation)
            reservation.service_probe()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during probe pending for reservation #{reservation.get_reservation_id()}",
                       e=e)

    def probe_pending_delegation(self, *, delegation: ABCDelegation):
        """
        Probes to check for completion of pending operation.
        @param delegation the delegation being probed
        @throws Exception rare
        """
        try:
            delegation.prepare_probe()
            delegation.probe_pending()
            self.plugin.get_database().update_delegation(delegation=delegation)
            delegation.service_probe()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during probe pending for delegation #{delegation.get_delegation_id()}",
                       e=e)

    def purge(self):
        """
        Purges all closed reservations.
        @throws Exception
        """
        for reservation in self.reservations.values():
            if reservation.is_closed():
                try:
                    reservation.get_kernel_slice().unregister(reservation=reservation)
                except Exception as e:
                    self.logger.error(f"An error occurred during purge for "
                                      f"reservation #{reservation.get_reservation_id()} e: {e}")
                finally:
                    self.reservations.remove(reservation=reservation)

        try:
            self.lock.acquire()
            delegations_to_be_removed = []
            for delegation in self.delegations.values():
                if delegation.is_closed():
                    try:
                        delegation.get_slice_object().unregister_delegation(delegation=delegation)
                    except Exception as e:
                        self.logger.error(f"An error occurred during purge for "
                                          f"delegation #{delegation.get_delegation_id()} e:{e}")
                    finally:
                        delegations_to_be_removed.append(delegation.get_delegation_id())

            for d in delegations_to_be_removed:
                if d in self.delegations:
                    self.delegations.pop(d)
        finally:
            self.lock.release()

    def query(self, *, properties: dict):
        """
        Processes a query request.
        @param properties query
        @return query response
        """
        return self.policy.query(p=properties)

    def redeem(self, *, reservation: ABCKernelControllerReservationMixin):
        """
        Redeem a reservation
        @param reservation reservation
        """
        try:
            if reservation.can_redeem():
                reservation.reserve(policy=self.policy)
            else:
                raise KernelException("The current reservation state prevent it from being redeemed")

            self.plugin.get_database().update_reservation(reservation=reservation)
            if not reservation.is_failed():
                reservation.service_reserve()
        except Exception as e:
            self.logger.error(f"An error occurred during redeem for reservation #{reservation.get_reservation_id()} "
                              f"e: {e}")

    def register(self, *, reservation: ABCKernelReservation, slice_object: ABCKernelSlice) -> bool:
        """
        Registers a new reservation with its slice and the kernel reservation
        table. Must be called with the kernel lock on.
        @param reservation local reservation object
        @param slice_object local slice object. The slice must have previously been
                   registered with the kernel.
        @return true if the reservation was registered, false otherwise
        @throws Exception
        """
        add = False
        reservation.set_logger(logger=self.logger)

        if not reservation.is_closed():
            # Note: as of now slice.register must be the first operation in
            # this method. slice.register will throw an exception if the
            # reservation is already present in the slice table.

            # register with the local slice
            slice_object.register(reservation=reservation)

            # register with the reservations table
            if self.reservations.contains(reservation=reservation):
                slice_object.unregister(reservation=reservation)
                raise KernelException("There is already a reservation with the given identifier")

            self.reservations.add(reservation=reservation)

            # attach actor to the reservation
            reservation.set_actor(actor=self.plugin.get_actor())
            # attach the local slice object
            reservation.set_slice(slice_object=slice_object)
            add = True
        else:
            self.logger.warning(f"Attempting to register a closed reservation #{reservation.get_reservation_id()}")

        return add

    def register_delegation_with_slice(self, *, delegation: ABCDelegation, slice_object: ABCKernelSlice) -> bool:
        """
        Registers a new delegation with its slice and the kernel delegation
        table. Must be called with the kernel lock on.
        @param delegation local delegation object
        @param slice_object local slice object. The slice must have previously been
                   registered with the kernel.
        @return true if the delegation was registered, false otherwise
        @throws Exception
        """
        add = False

        if not delegation.is_closed():
            # Note: as of now slice.register must be the first operation in
            # this method. slice.register will throw an exception if the
            # delegation is already present in the slice table.

            # register with the local slice
            slice_object.register_delegation(delegation=delegation)

            try:
                self.lock.acquire()
                # register with the delegation table
                if delegation.get_delegation_id() in self.delegations:
                    slice_object.unregister_delegation(delegation=delegation)
                    raise KernelException("There is already a delegation with the given identifier")

                self.delegations[delegation.get_delegation_id()] = delegation
            finally:
                self.lock.release()

            # attach actor to the reservation
            delegation.set_actor(actor=self.plugin.get_actor())
            # attach the local slice object
            delegation.set_slice_object(slice_object=slice_object)
            add = True
        else:
            self.logger.warning(f"Attempting to register a closed delegation #{delegation.get_delegation_id()}")

        return add

    def register_delegation(self, *, delegation: ABCDelegation):
        """
        Re-registers the delegation.
        @param delegation delegation
        @throws Exception
        """
        if delegation is None or delegation.get_delegation_id() is None or \
                delegation.get_slice_object() is None or delegation.get_slice_object().get_name() is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        local_slice = None
        add = False

        local_slice = self.slices.get(slice_id=delegation.get_slice_id(), raise_exception=True)
        add = self.register_delegation_with_slice(delegation=delegation, slice_object=local_slice)

        if add:
            try:
                self.plugin.get_database().add_delegation(delegation=delegation)
            except Exception as e:
                self.unregister_no_check_d(delegation=delegation, slice_object=local_slice)
                raise e

    def register_reservation(self, *, reservation: ABCKernelReservation):
        """
        Re-registers the reservation.
        @param reservation reservation
        @throws Exception
        """
        if reservation is None or reservation.get_reservation_id() is None or \
                reservation.get_slice() is None or reservation.get_slice().get_name() is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        local_slice = None
        add = False

        local_slice = self.slices.get(slice_id=reservation.get_slice().get_slice_id(), raise_exception=True)
        add = self.register(reservation=reservation, slice_object=local_slice)

        if add:
            try:
                self.plugin.get_database().add_reservation(reservation=reservation)
            except Exception as e:
                self.unregister_no_check(reservation=reservation, slice_object=local_slice)
                raise e

    def register_slice(self, *, slice_object: ABCKernelSlice):
        """
        Registers the specified slice with the kernel.
        @param slice_object slice to register
        @throws Exception if the slice cannot be registered
        """
        slice_object.prepare()
        self.slices.add(slice_object=slice_object)

        try:
            self.plugin.get_database().add_slice(slice_object=slice_object)
        except Exception as e:
            self.slices.remove(slice_id=slice_object.get_slice_id())
            self.logger.error(traceback.format_exc())
            self.error(err="could not register slice", e=e)

    def remove_reservation(self, *, rid: ID):
        """
        Removes the reservation.
        @param rid reservation id.
        @throws Exception
        """
        if rid is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        real = self.reservations.get(rid=rid)

        if real is not None:
            if real.is_closed() or real.is_failed() or real.get_state() == ReservationStates.CloseWait:
                self.unregister_reservation(rid=rid)
            else:
                raise KernelException("Only reservations in failed, closed, or closewait state can be removed.")
        else:
            self.logger.debug(f"Reservation # {rid} not found")

        self.plugin.get_database().remove_reservation(rid=rid)
        self.logger.debug(f"Reservation # {rid} removed from DB")

    def remove_slice(self, *, slice_id: ID):
        """
        Removes the specified slice.
        @param slice_id slice identifier
        @throws Exception if the slice contains active reservations or removal fails
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        possible = False
        slice_object = self.get_slice(slice_id=slice_id)

        if slice_object is None:
            self.logger.debug("Slice object not found in local data structure, removing from database")
            self.plugin.get_database().remove_slice(slice_id=slice_id)
        else:
            possible = (slice_object.get_reservations().size() == 0)

            if possible:
                # remove the slice from the slices table
                self.slices.remove(slice_id=slice_id)
                # release any resources assigned to the slice: unlocked,
                # because it may be blocking. The plugin is responsible for
                # synchronization.
                self.plugin.release_slice(slice_obj=slice_object)
                # remove from the database
                self.plugin.get_database().remove_slice(slice_id=slice_id)

    def re_register_delegation(self, *, delegation: ABCDelegation):
        """
        Re-registers the delegation.
        @param delegation delegation
        @throws Exception
        """
        if delegation is None or delegation.get_delegation_id() is None or \
                delegation.get_slice_object() is None or delegation.get_slice_object().get_slice_id() is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        local_slice = None
        local_slice = self.slices.get(slice_id=delegation.get_slice_object().get_slice_id(), raise_exception=True)

        if local_slice is None:
            raise KernelException("slice not registered with the kernel")
        else:
            self.register_delegation_with_slice(delegation=delegation, slice_object=local_slice)

        # Check if the delegation has a database record.
        temp = self.plugin.get_database().get_delegation(dlg_graph_id=delegation.get_delegation_id())

        if temp is None:
            self.unregister_no_check_d(delegation=delegation, slice_object=local_slice)
            raise KernelException("The delegation has no database record")

    def re_register_reservation(self, *, reservation: ABCKernelReservation):
        """
        Re-registers the reservation.
        @param reservation reservation
        @throws Exception
        """
        if reservation is None or reservation.get_reservation_id() is None or \
                reservation.get_slice() is None or reservation.get_slice().get_slice_id() is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        local_slice = None
        local_slice = self.slices.get(slice_id=reservation.get_slice().get_slice_id(), raise_exception=True)

        if local_slice is None:
            raise KernelException("slice not registered with the kernel")
        else:
            self.register(reservation=reservation, slice_object=local_slice)

        # Check if the reservation has a database record.
        temp = None
        temp = self.plugin.get_database().get_reservation(rid=reservation.get_reservation_id())

        if temp is None:
            self.unregister_no_check(reservation=reservation, slice_object=local_slice)
            raise KernelException("The reservation has no database record")

    def re_register_slice(self, *, slice_object: ABCKernelSlice):
        """
        Re-registers the specified slice with the kernel.
        @param slice_object slice to re-register
        @throws Exception if the slice cannot be registered
        """
        slice_object.prepare()
        self.slices.add(slice_object=slice_object)

        try:
            temp = self.plugin.get_database().get_slice(slice_id=slice_object.get_slice_id())
            if temp is None:
                raise KernelException("The slice does not have a database record")
        except Exception as e:
            self.slices.remove(slice_id=slice_object.get_slice_id())
            self.logger.error(traceback.format_exc())
            self.error(err="could not re-register slice", e=e)

    def reserve(self, *, reservation: ABCKernelReservation):
        """
        Handles a reserve operation for the reservation.
        Client: issue a ticket request or a claim request.
        Broker: process a request for a new ticket. Claims for previously
        exported tickets are handled by #claim(BrokerReservation).
        Authority: process a request for a new lease.
        @param reservation reservation for which to perform redeem
        @throws Exception
        """
        try:
            reservation.reserve(policy=self.policy)
            self.plugin.get_database().update_reservation(reservation=reservation)
            if not reservation.is_failed():
                reservation.service_reserve()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during reserve for reservation #{reservation.get_reservation_id()}", e=e)

    def delegate(self, *, delegation: ABCDelegation, id_token: str = None):
        """
        Handles a delegate operation for the delegation.
        Broker: process a request for a new delegate.
        Authority: process a request for a new delegate.
        @param delegation delegation
        @param id_token id token
        @throws Exception
        """
        try:
            delegation.delegate(policy=self.policy, id_token=id_token)
            self.plugin.get_database().update_delegation(delegation=delegation)
            if not delegation.is_closed():
                delegation.service_delegate()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during delegate for delegation #{delegation.get_delegation_id()}", e=e)

    def soft_validate_delegation(self, *, delegation: ABCDelegation = None, did: str = None) -> ABCDelegation:
        """
        Retrieves the locally registered delegation that corresponds to the
        passed delegation. Obtains the reservation from the containing slice
        object.
        @param delegation delegation being validated
        @param did delegation identifier or reservation being validated
        @return the locally registered did that corresponds to the passed
                delegation or null if no local did exists
        @throws Exception if the slice referenced by the incoming delegation is
                    not locally registered
        """
        if delegation is None and did is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        result = None

        if delegation is not None:
            # Each local delegation is indexed in two places: (1) The delegation
            # set in the kernel, and (2) inside the local slice object. Here we
            # will check to see if the local slice exists and will retrieve the
            # delegation from the local slice.
            s = self.get_local_slice(slice_object=delegation.get_slice_object())

            result = s.soft_lookup_delegation(did=delegation.get_delegation_id())

        if did is not None:
            try:
                self.lock.acquire()
                result = self.delegations.get(did, None)
            finally:
                self.lock.release()

        return result

    def soft_validate(self, *, reservation: ABCKernelReservation = None, rid: ID = None) -> ABCKernelReservation:
        """
        Retrieves the locally registered reservation that corresponds to the
        passed reservation. Obtains the reservation from the containing slice
        object.
        @param reservation reservation being validated
        @param rid reservation identifier or reservation being validated
        @return the locally registered reservation that corresponds to the passed
                reservation or null if no local reservation exists
        @throws IllegalArgumentException if the arguments are invalid
        @throws Exception if the slice referenced by the incoming reservation is
                    not locally registered
        """
        if reservation is None and rid is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        result = None

        if reservation is not None:
            # Each local reservation is indexed in two places: (1) The reservation
            # set in the kernel, and (2) inside the local slice object. Here we
            # will check to see if the local slice exists and will retrieve the
            # reservation from the local slice.
            s = self.get_local_slice(slice_object=reservation.get_slice())

            result = s.soft_lookup(rid=reservation.get_reservation_id())

        if rid is not None:
            result = self.reservations.get(rid=rid)

        return result

    def tick(self):
        """
        Timer interrupt.
        @throws Exception
        """
        try:
            try:
                self.lock.acquire()
                for delegation in self.delegations.values():
                    self.probe_pending_delegation(delegation=delegation)
            finally:
                self.lock.release()

            for reservation in self.reservations.values():
                self.probe_pending(reservation=reservation)

            try:
                self.lock.acquire()
                for slice_obj in self.slices.get_client_slices():
                    self.probe_pending_slices(slice_obj=slice_obj)
            finally:
                self.lock.release()

            self.purge()
            self.check_nothing_pending()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err="exception in Kernel.tick", e=e)

    def has_something_pending(self) -> bool:
        """
        Check is kernel has any pending reservations
        @return true if no terminal/nascent/pending reservations exist; false otherwise
        """
        for reservation in self.reservations.values():
            if not reservation.is_terminal() and (reservation.is_nascent() or not reservation.is_no_pending()):
                return True

        return False

    def check_nothing_pending(self):
        """
        Check for pending reservations
        """
        if not self.has_something_pending():
            with self.nothing_pending:
                self.nothing_pending.notify_all()

    def await_nothing_pending(self):
        """
        Await until nothing is pending
        """
        with self.nothing_pending:
            self.nothing_pending.wait()

    def unregister(self, *, reservation: ABCKernelReservation, slice_object: ABCKernelSlice):
        """
        Unregisters a reservation from the kernel data structures. Must be called
        with the kernel lock on. Performs state checks.
        @param reservation reservation to unregister
        @param slice_object local slice object
        @throws Exception
        """
        if reservation.is_closed() or reservation.is_failed() or reservation.get_state() == ReservationStates.CloseWait:
            slice_object.unregister(reservation=reservation)
            self.reservations.remove(reservation=reservation)
        else:
            raise KernelException("Only reservations in failed, closed, or closewait state can be unregistered.")

    def unregister_no_check(self, *, reservation: ABCKernelReservation, slice_object: ABCKernelSlice):
        """
        Unregisters a reservation from the kernel data structures. Must be called
        with the kernel lock on. Does not perform state checks.
        @param reservation reservation to unregister
        @param slice_object local slice object
        @throws Exception
        """
        slice_object.unregister(reservation=reservation)
        self.reservations.remove(reservation=reservation)

    def unregister_no_check_d(self, *, delegation: ABCDelegation, slice_object: ABCKernelSlice):
        """
        Unregisters a delegation from the kernel data structures. Must be called
        with the kernel lock on. Does not perform state checks.
        @param delegation delegation to unregister
        @param slice_object local slice object
        @throws Exception
        """
        slice_object.unregister_delegation(delegation=delegation)
        try:
            self.lock.acquire()
            if delegation.get_delegation_id() in self.delegations:
                self.delegations.pop(delegation.get_delegation_id())
        finally:
            if self.lock.locked():
                self.lock.release()

    def unregister_reservation(self, *, rid: ID):
        """
        Unregisters the reservation
        @param rid reservation id
        @throws Exception
        """
        if rid is None:
            raise KernelException(Constants.INVALID_ARGUMENT)
        local_reservation = self.reservations.get_exception(rid=rid)
        self.unregister(reservation=local_reservation, slice_object=local_reservation.get_kernel_slice())

        self.policy.remove(reservation=local_reservation)

    def unregister_slice(self, *, slice_id: ID):
        """
        Unregisters the specified slice.
        @param slice_id slice id
        @throws Exception if the slice cannot be unregistered (it has active
                    reservations) or has not been previously registered with the
                    kernel
        """
        if slice_id is None:
            raise KernelException(Constants.INVALID_ARGUMENT)

        s = self.get_slice(slice_id=slice_id)

        if s is None:
            raise KernelException("Trying to unregister a slice, which is not registered with the kernel")

        if s.get_reservations().size() == 0:
            self.slices.remove(slice_id=slice_id)
            self.plugin.release_slice(slice_obj=s)
        else:
            raise KernelException("Slice cannot be unregistered: not empty")

    def update_lease(self, *, reservation: ABCKernelReservation, update: Reservation, update_data: UpdateData):
        """
        Handles an incoming update lease operation (client side only).
        @param reservation local reservation
        @param update update sent from site authority
        @param update_data status of the operation the authority is informing us about
        @throws Exception
        """
        try:
            self.logger.debug(f"update_lease: Incoming term {update.get_term()}")
            reservation.update_lease(incoming=update, update_data=update_data)

            # NOTE: the database update has to happen BEFORE the service
            # update: we need to record the fact that we received a concrete
            # set, so that on recovery we can go and recover the concrete set.
            # If the database update is after the service call, we may end up
            # in Ticketed, Redeeming with state in the database that will
            # prevent us from incorporating a leaseUpdate.

            self.plugin.get_database().update_reservation(reservation=reservation)

            if not reservation.is_failed():
                reservation.service_update_lease()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during update lease for reservation "
                           f"# {reservation.get_reservation_id()}", e=e)

    def update_ticket(self, *, reservation: ABCKernelReservation, update: Reservation, update_data: UpdateData):
        """
        Handles an incoming update ticket operation (client side only).
        @param reservation local reservation
        @param update update sent from upstream broker
        @param update_data status of the operation the broker is informing us about
        @throws Exception
        """
        try:
            reservation.update_ticket(incoming=update, update_data=update_data)
            self.plugin.get_database().update_reservation(reservation=reservation)
            if not reservation.is_failed():
                reservation.service_update_ticket()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during update ticket for "
                           f"reservation # {reservation.get_reservation_id()}", e=e)

    def update_delegation(self, *, delegation: ABCDelegation, update: ABCDelegation, update_data: UpdateData):
        """
        Handles an incoming update delegation operation (client side only).
        @param delegation local delegation
        @param update update sent from upstream broker
        @param update_data status of the operation the broker is informing us about
        @throws Exception
        """
        try:
            delegation.update_delegation(incoming=update, update_data=update_data)
            self.plugin.get_database().update_delegation(delegation=delegation)
            delegation.service_update_delegation()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.error(err=f"An error occurred during update delegation for "
                           f"delegation # {delegation.get_delegation_id()}", e=e)

    def validate(self, *, reservation: ABCKernelReservation = None, rid: ID = None):
        """
        Retrieves the locally registered reservation that corresponds to the
        passed reservation. Obtains the reservation from the containing slice
        object.
        @param reservation reservation being validated
        @param rid reservation id
        @return the locally registered reservation that corresponds to the passed
                reservation
        @throws Exception if there is no local reservation that corresponds to
                    the passed reservation
        """
        if (reservation is not None and rid is not None) or (reservation is None and rid is None):
            raise KernelException(Constants.INVALID_ARGUMENT)

        if reservation is not None:
            local = self.soft_validate(reservation=reservation)
            if local is None:
                self.error(err="reservation not found", e=ReservationNotFoundException(rid=rid))
            return local

        if rid is not None:
            local = self.soft_validate(rid=rid)
            if local is None:
                self.error(err="reservation not found", e=ReservationNotFoundException(rid=rid))
            return local

        return None

    def handle_failed_rpc(self, *, reservation: ABCKernelReservation, rpc: FailedRPC):
        """
        Handle failed rpc
        @param reservation reservation
        @param rpc rpc
        """
        reservation.handle_failed_rpc(failed=rpc)

    def validate_delegation(self, *, delegation: ABCDelegation = None, did: str = None):
        """
        Retrieves the locally registered delegation that corresponds to the
        passed delegation. Obtains the delegation from the containing slice
        object.
        @param delegation delegation being validated
        @param did delegation id
        @return the locally registered delegation that corresponds to the passed
                delegation
        @throws Exception if there is no local delegation that corresponds to
                    the passed delegation
        """
        if (delegation is not None and did is not None) or (delegation is None and did is None):
            raise KernelException(Constants.INVALID_ARGUMENT)

        if delegation is not None:
            local = self.soft_validate_delegation(delegation=delegation)
            if local is None:
                self.error(err="delegation not found", e=DelegationNotFoundException(did=did))
            return local

        if did is not None:
            local = self.soft_validate_delegation(did=did)
            if local is None:
                self.error(err="delegation not found", e=DelegationNotFoundException(did=did))
            return local

        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric_mb.message_bus.messages.close_avro import CloseAvro
from fabric_mb.message_bus.messages.extend_lease_avro import ExtendLeaseAvro
from fabric_mb.message_bus.messages.modify_lease_avro import ModifyLeaseAvro
from fabric_mb.message_bus.messages.redeem_avro import RedeemAvro
from fabric_mb.message_bus.messages.reservation_avro import ReservationAvro
from fabric_mb.message_bus.messages.message import IMessageAvro

from fabric_cf.actor.core.common.exceptions import ProxyException
from fabric_cf.actor.core.kernel.authority_reservation import AuthorityReservationFactory
from fabric_cf.actor.core.kernel.incoming_reservation_rpc import IncomingReservationRPC
from fabric_cf.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric_cf.actor.core.proxies.kafka.translate import Translate
from fabric_cf.actor.core.proxies.kafka.services.broker_service import BrokerService
from fabric_cf.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_authority_reservation import ABCAuthorityReservation


class AuthorityService(BrokerService):

    def pass_authority(self, *, reservation: ReservationAvro) -> ABCAuthorityReservation:
        slice_obj = Translate.translate_slice(slice_avro=reservation.slice)
        term = Translate.translate_term_from_avro(term=reservation.term)

        resource_set = Translate.translate_resource_set_from_avro(rset=reservation.resource_set)
        cset = self.get_concrete(reservation=reservation)
        if cset is None:
            raise ProxyException("Unsupported Concrete type")

        resource_set.set_resources(cset=cset)
        rid = ID(uid=reservation.reservation_id)

        result = AuthorityReservationFactory.create(resources=resource_set, term=term, slice_obj=slice_obj, rid=rid,
                                                    actor=self.actor)
        result.set_owner(owner=self.actor.get_identity())
        result.set_sequence_in(sequence=reservation.sequence)

        return result

    def close(self, *, request: CloseAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            rsvn = self.pass_authority(reservation=request.reservation)
            rpc = IncomingReservationRPC(message_id=ID(uid=request.message_id), request_type=RPCRequestType.Close,
                                         reservation=rsvn, caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid close request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def redeem(self, *, request: RedeemAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            rsvn = self.pass_authority(reservation=request.reservation)
            callback = self.get_callback(kafka_topic=request.callback_topic, auth=auth_token)
            rpc = IncomingReservationRPC(message_id=ID(uid=request.message_id), request_type=RPCRequestType.Redeem,
                                         reservation=rsvn, callback=callback, caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid redeem request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def extend_lease(self, *, request: ExtendLeaseAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            rsvn = self.pass_authority(reservation=request.reservation)
            rpc = IncomingReservationRPC(message_id=ID(uid=request.message_id),
                                         request_type=RPCRequestType.ExtendLease, reservation=rsvn, caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid extend_lease request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def modify_lease(self, *, request: ModifyLeaseAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            rsvn = self.pass_authority(reservation=request.reservation)
            rpc = IncomingReservationRPC(message_id=ID(uid=request.message_id),
                                         request_type=RPCRequestType.ModifyLease, reservation=rsvn, caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid modify_lease request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def process(self, *, message: IMessageAvro):
        if message.get_message_name() == IMessageAvro.close:
            self.close(request=message)
        elif message.get_message_name() == IMessageAvro.redeem:
            self.redeem(request=message)
        elif message.get_message_name() == IMessageAvro.extend_lease:
            self.extend_lease(request=message)
        elif message.get_message_name() == IMessageAvro.modify_lease:
            self.modify_lease(request=message)
        elif message.get_message_name() == IMessageAvro.result_reservation:
            self.logger.debug("Claim Resources Response receieved: {}".format(message))
        elif message.get_message_name() == IMessageAvro.result_delegation:
            self.logger.debug("Claim Delegation Response receieved: {}".format(message))
        else:
            super().process(message=message)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import queue
import threading
import traceback
from datetime import datetime
import time

from fim.user import ServiceType

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates
from fabric_cf.actor.core.time.term import Term
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_slice_wrapper import OrchestratorSliceWrapper


class SliceDeferThread:
    """
    This runs as a standalone thread started by Orchestrator and deals with slices that have
    NetworkService reservations which have to wait for NetworkNode reservations to be ticketed
    """
    THREAD_SLEEP_TIME_IN_SECONDS = 5
    DEFAULT_MAX_CREATE_TIME_IN_SECONDS = 600

    def __init__(self):
        self.deferred_slices = queue.Queue()
        self.defer_slice_avail_condition = threading.Condition()
        self.thread_lock = threading.Lock()
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()

        self.thread = None
        self.last_slice = None
        self.last_slice_time = None
        self.max_create_wait_time = 0
        self.stopped = False

        wait_time = GlobalsSingleton.get().get_config().get_runtime_config().get(
            Constants.PROPERTY_CONF_CONTROLLER_CREATE_WAIT_TIME, None)

        if wait_time is None:
            self.max_create_wait_time = self.DEFAULT_MAX_CREATE_TIME_IN_SECONDS
        else:
            self.max_create_wait_time = wait_time

        self.delay_resource_types = []
        self.delay_resource_types.append(str(ServiceType.L2STS))
        self.delay_resource_types.append(str(ServiceType.L2Bridge))
        self.delay_resource_types.append(str(ServiceType.L2PTP))

    def queue_slice(self, *, controller_slice: OrchestratorSliceWrapper):
        """
        Queue a slice
        :param controller_slice:
        :return:
        """
        with self.defer_slice_avail_condition:
            self.deferred_slices.put_nowait(controller_slice)
            self.logger.debug(f"Added slice to deferred slices queue {controller_slice.__class__.__name__}")
            time.sleep(self.THREAD_SLEEP_TIME_IN_SECONDS)
            self.defer_slice_avail_condition.notify_all()

    def update_last(self, *, controller_slice: OrchestratorSliceWrapper):
        """
        Update the last slice
        :param controller_slice:
        :return:
        """
        if controller_slice is None:
            return

        self.logger.info(f"Updating last slice with: {controller_slice.get_slice_name()}/"
                         f"{controller_slice.get_slice_id()}")
        self.last_slice = controller_slice
        self.last_slice_time = datetime.utcnow()

    def process_slice(self, *, controller_slice: OrchestratorSliceWrapper):
        """
        Process a slice
        :param controller_slice:
        :return:
        """
        if controller_slice is None:
            return

        self.logger.info(f"Processing slice {controller_slice.get_slice_name()}/{controller_slice.get_slice_id()} "
                         f"immediately")
        if self.check_computed_reservations(controller_slice=controller_slice):
            self.update_last(controller_slice=controller_slice)

        if self.demand_slice(controller_slice=controller_slice):
            # Slice has pending reservations
            self.queue_slice(controller_slice=controller_slice)

    def start(self):
        """
        Start thread
        :return:
        """
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise OrchestratorException("This SliceDeferThread has already been started")

            self.thread = threading.Thread(target=self.run)
            self.thread.setName(self.__class__.__name__)
            self.thread.setDaemon(True)
            self.thread.start()

        finally:
            self.thread_lock.release()

    def stop(self):
        """
        Stop thread
        :return:
        """
        self.stopped = True
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning("It seems that the SliceDeferThread is running. Interrupting it")
                try:
                    # TODO find equivalent of interrupt
                    with self.defer_slice_avail_condition:
                        self.defer_slice_avail_condition.notify_all()
                    temp.join()
                except Exception as e:
                    self.logger.error(f"Could not join SliceDeferThread thread {e}")
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def run(self):
        """
        Thread main loop
        :return:
        """
        self.logger.debug("SliceDeferThread started")
        while True:
            if self.last_slice is not None:
                self.logger.info(f"Processing previously deferred slice {self.last_slice.get_slice_name()}/"
                                 f"{self.last_slice.get_slice_id()}")

                try:
                    self.last_slice.lock()
                    if self.delay_not_done(controller_slice=self.last_slice):
                        current_time = datetime.utcnow()
                        if (current_time - self.last_slice_time).seconds > self.max_create_wait_time:
                            self.logger.info(f"Maximum wait time exceeded for slice: {self.last_slice.get_slice_name()}/"
                                             f"{self.last_slice.get_slice_id()}, proceeding anyway")
                            # This shall trigger failure
                            self.demand_slice(controller_slice=self.last_slice, force=True)
                        else:
                            continue
                except Exception as e:
                    self.logger.error(f"Exception while checking slice {self.last_slice.get_slice_name()}/"
                                      f"{self.last_slice.get_slice_id()} e: {e}")
                    self.logger.error(traceback.format_exc())
                finally:
                    self.last_slice.unlock()

            controller_slice = None

            with self.defer_slice_avail_condition:

                while self.deferred_slices.empty() and not self.stopped:
                    try:
                        self.defer_slice_avail_condition.wait()
                    except InterruptedError as e:
                        self.logger.error(e)
                        self.logger.info("SliceDeferThread thread interrupted. Exiting")
                        return

                if self.stopped:
                    self.logger.info("SliceDeferThread exiting")
                    return

                if not self.deferred_slices.empty():
                    controller_slice = self.deferred_slices.get_nowait()

                self.defer_slice_avail_condition.notify_all()

            if controller_slice is None:
                continue

            self.logger.info(f"Performing demand on deferred slice {controller_slice.get_slice_name()}/"
                             f"{controller_slice.get_slice_id()}")

            self.update_last(controller_slice=controller_slice)
            try:
                controller_slice.lock()
                if self.demand_slice(controller_slice=controller_slice):
                    self.logger.debug("Adding slice back to the queue!")
                    self.queue_slice(controller_slice=controller_slice)
            except Exception as e:
                self.logger.error(f"Exception while demanding slice: {self.last_slice.get_slice_name()}/"
                                  f"{self.last_slice.get_slice_id()} e: {e}")
            finally:
                controller_slice.unlock()

        self.logger.debug("SliceDeferThread exited")

    def demand_slice(self, *, controller_slice: OrchestratorSliceWrapper, force: bool = False) -> bool:
        """
        Demand slice reservations. If any of the reservations have predecessors which are not ticketed yet;
        those reservations are not demanded yet i.e. no Ticket request to broker is triggered for them.
        :param controller_slice:
        :param force:
        :return: True if slice should be added to deferred queue to demand the delayed reservations; False otherwise
        """
        ret_val = False
        if controller_slice is None:
            self.logger.error("demand slice was given a None slice")
            return ret_val

        computed_reservations = controller_slice.get_computed_reservations()

        if computed_reservations is None:
            return ret_val

        try:
            from fabric_cf.orchestrator.core.orchestrator_kernel import OrchestratorKernelSingleton
            controller = OrchestratorKernelSingleton.get().get_management_actor()
            for reservation in computed_reservations:
                if reservation.get_reservation_id() in controller_slice.demanded_reservations():
                    self.logger.debug(f"Reservation: {reservation.get_reservation_id()} already demanded")
                    continue

                self.logger.debug(f"Issuing demand for reservation: {reservation.get_reservation_id()}")

                if reservation.get_state() != ReservationStates.Unknown.value:
                    self.logger.debug(f"Reservation not in {reservation.get_state()} state, ignoring it")
                    continue

                if not controller_slice.check_predecessors_ticketed(reservation=reservation) and not force:
                    self.logger.info(f"Reservation waiting for predecessors to be ticketed, ignoring it")
                    ret_val = True
                    continue

                if not controller.demand_reservation(reservation=reservation):
                    raise OrchestratorException(f"Could not demand resources: {controller.get_last_error()}")
                controller_slice.mark_demanded(rid=reservation.get_reservation_id())
                self.logger.debug(f"Reservation #{reservation.get_reservation_id()} demanded successfully")
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Unable to get orchestrator or demand reservation: {}".format(e))

        return ret_val

    def check_computed_reservations(self, *, controller_slice: OrchestratorSliceWrapper) -> bool:
        """
        Check computed reservations to see if there are any network sliver reservations
        :param controller_slice:
        :return: True if slice has network sliver reservations
        """
        for r in controller_slice.get_computed_reservations():
            if r.get_resource_type() in self.delay_resource_types:
                self.logger.info(f"Slice: {controller_slice.get_slice_name()}/{controller_slice.get_slice_id()} "
                                 f" has delayed resource type: {r.get_resource_type()}")
            return True

        return False

    def delay_not_done(self, *, controller_slice: OrchestratorSliceWrapper) -> bool:
        """
        Check if predecessors for network sliver reservations have been ticketed
        :param controller_slice:
        :return: True if Predecessors for Network Sliver Reservations have not been ticketed; False otherwise
        """
        if controller_slice is None:
            return False

        self.logger.info("Checking slice {}/{}".format(controller_slice.get_slice_name(),
                                                       controller_slice.get_slice_id()))

        all_reservations = None

        try:
            all_reservations = controller_slice.get_all_reservations()
        except Exception as e:
            self.logger.error(f"Exception in delay_not_done for slice: {controller_slice.get_slice_name()}/"
                              f"{controller_slice.get_slice_id()} e: {e}")

        if all_reservations is None:
            self.logger.info(f"Slice: {controller_slice.get_slice_name()}/"
                             f"{controller_slice.get_slice_id()} has None reservations in delay_not_done")
            # slice has not been submitted
            return self.check_computed_reservations(controller_slice=controller_slice)
        else:
            if len(all_reservations) <= 0:
                self.logger.info(f"Slice: {controller_slice.get_slice_name()}/{controller_slice.get_slice_id()} "
                                 f"has empty reservations in delay_not_done")
                # slice has not been submitted
                return self.check_computed_reservations(controller_slice=controller_slice)

            for reservation in all_reservations:
                if reservation.get_resource_type() in self.delay_resource_types:
                    # Reservation is Network Sliver Reservation;
                    # Reservation is not in a final state, we need to wait
                    reservation_state = ReservationStates(reservation.get_state())
                    if reservation_state != ReservationStates.Active and \
                            reservation_state != ReservationStates.Closed and \
                            reservation_state != ReservationStates.CloseWait and \
                            reservation_state != ReservationStates.Failed:
                        self.logger.info(f"Slice: {controller_slice.get_slice_name()}/{controller_slice.get_slice_id()}"
                                         f" has resource type: {reservation.get_resource_type()}"
                                         f" with reservation: {reservation.get_reservation_id()} "
                                         f"in state: {reservation_state.name} "
                                         f"that is not yet done")

                        # Use reservation from computed reservation as reservation has not been demanded yet
                        if controller_slice.check_predecessors_ticketed_by_rid(rid=reservation.get_reservation_id()):
                            self.logger.info("Predecessors ticketed, reservation can be demanded now!")
                            return False

                        return True
            self.logger.info(f"Slice: {controller_slice.get_slice_name()}/{controller_slice.get_slice_id()} "
                             f"has no non-final reservations ({len(all_reservations)})")

            return False



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import sched
import sys
import threading
import traceback
from datetime import datetime, timedelta
from logging.handlers import RotatingFileHandler
from typing import TYPE_CHECKING

import logging
import os

from fss_utils.jwt_validate import JWTValidator

from fabric_cf.actor.core.common.exceptions import InitializationException
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.container.container import Container

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_actor_container import ABCActorContainer
    from fabric_cf.actor.boot.configuration import Configuration

logging.TRACE = 5
logging.addLevelName(logging.TRACE, "TRACE")
logging.Logger.trace = lambda inst, msg, *args, **kwargs: inst.log(logging.TRACE, msg, *args, **kwargs)
logging.trace = lambda msg, *args, **kwargs: logging.log(logging.TRACE, msg, *args, **kwargs)


class Globals:
    config_file = Constants.CONFIGURATION_FILE

    def __init__(self):
        self.config = None
        self.log = None
        self.initialized = False
        self.started = False
        self.start_completed = False
        self.container = None
        self.properties = None
        self.timer_scheduler = sched.scheduler()
        self.timer_thread = None
        self.timer_condition = threading.Condition()
        self.lock = threading.Lock()
        self.jwt_validator = None

    def make_logger(self, *, log_config: dict = None):
        """
        Detects the path and level for the log file from the actor config and sets
        up a logger. Instead of detecting the path and/or level from the
        config, a custom path and/or level for the log file can be passed as
        optional arguments.

       :param log_config: Log config
       :return: logging.Logger object
        """
        if log_config is None:
            if self.config is None:
                raise RuntimeError('No config information available')

            log_config = self.config.get_global_config().get_logging()
        if log_config is None:
            raise RuntimeError('No logging  config information available')

        log_path = None
        if Constants.PROPERTY_CONF_LOG_DIRECTORY in log_config and Constants.PROPERTY_CONF_LOG_FILE in log_config:
            log_path = log_config[Constants.PROPERTY_CONF_LOG_DIRECTORY] + '/' + \
                       log_config[Constants.PROPERTY_CONF_LOG_FILE]

        if log_path is None:
            raise RuntimeError('The log file path must be specified in config or passed as an argument')

        # Get the log level
        log_level = None
        if Constants.PROPERTY_CONF_LOG_LEVEL in log_config:
            log_level = log_config.get(Constants.PROPERTY_CONF_LOG_LEVEL, None)

        if log_level is None:
            log_level = logging.INFO

        # Set up the root logger
        log = logging.getLogger(log_config.get(Constants.PROPERTY_CONF_LOGGER, None))
        log.setLevel(log_level)
        log_format = \
            '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s] - %(levelname)s - %(message)s'

        os.makedirs(os.path.dirname(log_path), exist_ok=True)

        backup_count = log_config.get(Constants.PROPERTY_CONF_LOG_RETAIN, None)
        max_log_size = log_config.get(Constants.PROPERTY_CONF_LOG_SIZE, None)

        file_handler = RotatingFileHandler(log_path, backupCount=int(backup_count), maxBytes=int(max_log_size))

        logging.basicConfig(handlers=[file_handler], format=log_format)

        return log

    @staticmethod
    def delete_super_block():
        """
        Delete Super block file
        """
        if os.path.isfile(Constants.SUPERBLOCK_LOCATION):
            os.remove(Constants.SUPERBLOCK_LOCATION)

    def fail(self, *, e: Exception):
        """
        Fail the Actor
        @param e exception
        """
        self.log.error("Critical error: Actor failed to initialize {}".format(e))
        sys.exit(-1)

    def initialize(self):
        """
        Initialize the container and actor
        """
        try:
            self.lock.acquire()
            if not self.initialized:
                self.load_config()
                self.log = self.make_logger()
                self.log.info("Checking if connection to Kafka broker can be established")
                admin_kafka_client = self.get_kafka_admin_client()
                admin_kafka_client.list_topics()
                self.log.info("Connection to Kafka broker established successfully")
                self.load_jwt_validator()
                self.log.info("Main initialization complete.")
                self.initialized = True
        finally:
            self.lock.release()

    def load_jwt_validator(self):
        oauth_config = self.config.get_oauth_config()
        CREDMGR_CERTS = oauth_config.get(Constants.PROPERTY_CONF_O_AUTH_JWKS_URL, None)
        CREDMGR_KEY_REFRESH = oauth_config.get(Constants.PROPERTY_CONF_O_AUTH_KEY_REFRESH, None)
        self.log.info(f'Initializing JWT Validator to use {CREDMGR_CERTS} endpoint, '
                      f'refreshing keys every {CREDMGR_KEY_REFRESH} HH:MM:SS')
        t = datetime.strptime(CREDMGR_KEY_REFRESH, "%H:%M:%S")
        self.jwt_validator = JWTValidator(url=CREDMGR_CERTS,
                                          refresh_period=timedelta(hours=t.hour, minutes=t.minute, seconds=t.second))

    def load_config(self):
        """
        Load the configuration
        """
        try:
            from fabric_cf.actor.boot.configuration_loader import ConfigurationLoader
            loader = ConfigurationLoader(path=self.config_file)
            self.config = loader.read_configuration()
        except Exception as e:
            raise RuntimeError("Unable to parse configuration file {}".format(e))

    def get_jwt_validator(self):
        return self.jwt_validator

    def get_container(self) -> ABCActorContainer:
        """
        Get the container
        @return container
        """
        if not self.initialized:
            raise InitializationException(Constants.UNINITIALIZED_STATE)
        return self.container

    def get_config(self) -> Configuration:
        """
        Get the configuration
        @return config
        """
        if not self.initialized:
            raise InitializationException(Constants.UNINITIALIZED_STATE)
        return self.config

    def get_log_config(self) -> dict:
        """
        Get the Log configuration
        @return dict
        """
        return self.get_config().get_log_config()

    def get_kafka_config_admin_client(self) -> dict:
        """
        Get Kafka Config Admin Client
        @retun admin client config
        """
        if self.config is None or self.config.get_runtime_config() is None:
            return None
        bootstrap_server = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SERVER, None)
        security_protocol = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SECURITY_PROTOCOL, None)
        ssl_ca_location = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_S_SL_CA_LOCATION, None)
        ssl_certificate_location = self.config.get_runtime_config().get(
            Constants.PROPERTY_CONF_KAFKA_SSL_CERTIFICATE_LOCATION, None)
        ssl_key_location = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SSL_KEY_LOCATION, None)
        ssl_key_password = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SSL_KEY_PASSWORD, None)
        sasl_username = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_USERNAME, None)
        sasl_password = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_PASSWORD, None)
        sasl_mechanism = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SASL_MECHANISM, None)

        conf = {Constants.BOOTSTRAP_SERVERS: bootstrap_server,
                Constants.SECURITY_PROTOCOL: security_protocol,
                Constants.SSL_CA_LOCATION: ssl_ca_location,
                Constants.SSL_CERTIFICATE_LOCATION: ssl_certificate_location,
                Constants.SSL_KEY_LOCATION: ssl_key_location,
                Constants.SSL_KEY_PASSWORD: ssl_key_password}

        if sasl_username is not None and sasl_username != '' and sasl_password is not None and sasl_password != '':
            conf[Constants.SASL_USERNAME] = sasl_username
            conf[Constants.SASL_PASSWORD] = sasl_password
            conf[Constants.SASL_MECHANISM] = sasl_mechanism

        return conf

    def get_kafka_config_producer(self) -> dict:
        """
        Get Producer Config
        @return producer config
        """
        if self.config is None or self.config.get_runtime_config() is None:
            return None
        bootstrap_server = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SERVER, None)
        schema_registry = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SCHEMA_REGISTRY, None)
        security_protocol = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SECURITY_PROTOCOL, None)
        ssl_ca_location = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_S_SL_CA_LOCATION, None)
        ssl_certificate_location = self.config.get_runtime_config().get(
            Constants.PROPERTY_CONF_KAFKA_SSL_CERTIFICATE_LOCATION, None)
        ssl_key_location = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SSL_KEY_LOCATION, None)
        ssl_key_password = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SSL_KEY_PASSWORD, None)

        sasl_username = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_USERNAME, None)
        sasl_password = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_PASSWORD, None)
        sasl_mechanism = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SASL_MECHANISM, None)

        conf = {Constants.BOOTSTRAP_SERVERS: bootstrap_server,
                Constants.SECURITY_PROTOCOL: security_protocol,
                Constants.SSL_CA_LOCATION: ssl_ca_location,
                Constants.SSL_CERTIFICATE_LOCATION: ssl_certificate_location,
                Constants.SSL_KEY_LOCATION: ssl_key_location,
                Constants.SSL_KEY_PASSWORD: ssl_key_password,
                Constants.SCHEMA_REGISTRY_URL: schema_registry}

        if sasl_username is not None and sasl_username != '' and sasl_password is not None and sasl_password != '':
            conf[Constants.SASL_USERNAME] = sasl_username
            conf[Constants.SASL_PASSWORD] = sasl_password
            conf[Constants.SASL_MECHANISM] = sasl_mechanism

        return conf

    def get_kafka_config_consumer(self) -> dict:
        """
        Get Consumer config
        @return consumer config
        """
        if self.config is None or self.config.get_runtime_config() is None:
            return None
        conf = self.get_kafka_config_producer()

        group_id = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_GROUP_ID, None)

        conf['auto.offset.reset'] = 'earliest'

        sasl_username = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SASL_CONSUMER_USERNAME, None)
        sasl_password = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_SASL_CONSUMER_PASSWORD, None)

        if sasl_username is not None and sasl_username != '' and sasl_password is not None and sasl_password != '':
            conf[Constants.SASL_USERNAME] = sasl_username
            conf[Constants.SASL_PASSWORD] = sasl_password
        conf[Constants.GROUP_ID] = group_id
        return conf

    def get_kafka_schemas(self):
        """
        Get Avro schema
        @return key and value schema
        """
        key_schema_file = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_KEY_SCHEMA, None)
        value_schema_file = self.config.get_runtime_config().get(Constants.PROPERTY_CONF_KAFKA_VALUE_SCHEMA, None)

        from confluent_kafka import avro
        file = open(key_schema_file, "r")
        kbytes = file.read()
        file.close()
        key_schema = avro.loads(kbytes)
        file = open(value_schema_file, "r")
        vbytes = file.read()
        file.close()
        val_schema = avro.loads(vbytes)

        return key_schema, val_schema

    def get_kafka_producer(self):
        """
        Create and return a kafka producer
        @return producer
        """
        conf = self.get_kafka_config_producer()
        key_schema, val_schema = self.get_kafka_schemas()

        from fabric_mb.message_bus.producer import AvroProducerApi
        producer = AvroProducerApi(conf=conf, key_schema=key_schema, record_schema=val_schema, logger=self.get_logger())
        return producer

    def get_simple_kafka_producer(self):
        """
        Create and return a kafka producer
        @return producer
        """
        conf = self.get_kafka_config_producer()
        if conf is not None:
            conf.pop(Constants.SCHEMA_REGISTRY_URL)

        bqm_config = self.config.get_global_config().get_bqm_config()
        if bqm_config is not None:
            prod_user_name = bqm_config.get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_USERNAME, None)
            prod_password = bqm_config.get(Constants.PROPERTY_CONF_KAFKA_SASL_PRODUCER_PASSWORD, None)
            if prod_user_name is not None and prod_password is not None:
                conf[Constants.SASL_USERNAME] = prod_user_name
                conf[Constants.SASL_PASSWORD] = prod_password

        from confluent_kafka import Producer
        producer = Producer(conf)
        return producer

    def get_kafka_admin_client(self):
        """
        Create and return a kafka admin client
        @return admin client
        """
        from fabric_mb.message_bus.admin import AdminApi
        conf = self.get_kafka_config_admin_client()
        admin = AdminApi(conf=conf)
        return admin

    def get_logger(self):
        """
        Get logger
        @return logger
        """
        if not self.initialized:
            raise InitializationException(Constants.UNINITIALIZED_STATE)

        if self.log is None:
            self.log = self.make_logger()
        return self.log

    def start(self, *, force_fresh: bool):
        """
        Start CF Actor
        @param force_fresh true if clean restart, false if stateful restart
        """
        try:
            try:
                self.lock.acquire()
                if self.started:
                    return
                self.started = True
                self.start_completed = False
                if force_fresh:
                    self.delete_super_block()
            finally:
                self.lock.release()

            self.initialize()
            self.start_timer_thread()
            try:
                self.lock.acquire()
                self.container = Container()
                self.log.info("Successfully instantiated the container implementation.")
                self.log.info("Initializing container")
                self.container.initialize(config=self.config)
                self.log.info("Successfully initialized the container")
                self.start_completed = True
            finally:
                self.lock.release()
        except Exception as e:
            self.fail(e=e)

    def stop(self):
        """
        Stop the Actor
        """
        try:
            self.lock.acquire()
            if not self.started:
                return
            self.log.info("Stopping Actor")
            self.started = False
            self.stop_timer_thread()
            self.get_container().shutdown()
        except Exception as e:
            self.log.error("Error while shutting down: {}".format(e))
        finally:
            self.lock.release()

    def start_timer_thread(self):
        """
        Start the timer thread
        """
        if self.timer_thread is not None:
            raise RuntimeError("This timer thread has already been started")

        self.timer_thread = threading.Thread(target=self.timer_loop)
        self.timer_thread.setName('GlobalTimer')
        self.timer_thread.setDaemon(True)
        self.timer_thread.start()

    def stop_timer_thread(self):
        """
        Stop timer thread
        """
        temp = self.timer_thread
        self.timer_thread = None
        if temp is not None:
            self.log.warning("It seems that the timer thread is running. Interrupting it")
            try:
                with self.timer_condition:
                    self.timer_condition.notify_all()
                temp.join()
            except Exception as e:
                self.log.error("Could not join timer thread {}".format(e))

    def timer_loop(self):
        """
        Timer thread run function
        """
        self.log.debug("Timer thread started")
        while True:
            with self.timer_condition:
                while self.timer_scheduler.empty() and self.started:
                    try:
                        #self.log.debug("Waiting for condition")
                        self.timer_condition.wait()
                    except InterruptedError as e:
                        self.log.error(traceback.format_exc())
                        self.log.error("Timer thread interrupted. Exiting {}".format(e))
                        return

                    if not self.started:
                        self.log.info("Timer thread exiting")
                        return

                    self.timer_condition.notify_all()

                if not self.timer_scheduler.empty():
                    #self.log.debug("Executing Scheduled items")
                    self.timer_scheduler.run(blocking=False)


class GlobalsSingleton:
    """
    Global Singleton class
    """
    __instance = None

    def __init__(self):
        if self.__instance is not None:
            raise InitializationException("Singleton can't be created twice !")

    def get(self):
        """
        Actually create an instance
        """
        if self.__instance is None:
            self.__instance = Globals()
        return self.__instance

    get = classmethod(get)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_actor_mixin import ActorType
from fabric_cf.actor.core.common.constants import Constants, ErrorCodes
from fabric_cf.actor.core.manage.converter import Converter
from fabric_cf.actor.core.manage.management_object import ManagementObject
from fabric_cf.actor.core.manage.proxy_protocol_descriptor import ProxyProtocolDescriptor
from fabric_cf.actor.core.apis.abc_management_object import ABCManagementObject
from fabric_mb.message_bus.messages.result_proxy_avro import ResultProxyAvro
from fabric_cf.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric_cf.actor.core.util.id import ID
from fabric_mb.message_bus.messages.result_actor_avro import ResultActorAvro
from fabric_mb.message_bus.messages.result_avro import ResultAvro

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_container_database import ABCContainerDatabase
    from fabric_cf.actor.security.auth_token import AuthToken


class ContainerManagementObject(ManagementObject):
    def __init__(self):
        super().__init__()
        self.id = ID(uid=Constants.CONTAINER_MANAGMENT_OBJECT_ID)

    def register_protocols(self):
        from fabric_cf.actor.core.manage.local.local_container import LocalContainer
        local = ProxyProtocolDescriptor(protocol=Constants.PROTOCOL_LOCAL,
                                        proxy_class=LocalContainer.__name__,
                                        proxy_module=LocalContainer.__module__)

        from fabric_cf.actor.core.manage.kafka.kafka_container import KafkaContainer
        kakfa = ProxyProtocolDescriptor(protocol=Constants.PROTOCOL_KAFKA,
                                        proxy_class=KafkaContainer.__name__,
                                        proxy_module=KafkaContainer.__module__)

        self.proxies = []
        self.proxies.append(local)
        self.proxies.append(kakfa)

    def save(self) -> dict:
        properties = super().save()
        properties[Constants.PROPERTY_CLASS_NAME] = ContainerManagementObject.__name__
        properties[Constants.PROPERTY_MODULE_NAME] = ContainerManagementObject.__module__

        return properties

    @staticmethod
    def get_container_management_database() -> ABCContainerDatabase:
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        return GlobalsSingleton.get().get_container().get_database()

    @staticmethod
    def get_actors_from_registry(*, atype: int, caller: AuthToken):
        result = []
        actors = ActorRegistrySingleton.get().get_actors()
        if actors is not None:
            for a in actors:
                if atype == ActorType.All.value or atype == a.get_type():
                    result.append(a)
        return result

    def do_get_actors(self, *, atype: int, caller: AuthToken, id_token: str = None) -> ResultActorAvro:
        result = ResultActorAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            act_list = self.get_actors_from_registry(atype=atype, caller=caller)
            result.actors = Converter.fill_actors(act_list=act_list)
        except Exception as e:
            self.logger.error("get_actors {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_actors(self, *, caller: AuthToken, id_token: str = None) -> ResultActorAvro:
        return self.do_get_actors(atype=ActorType.All.value, caller=caller, id_token=id_token)

    def do_get_actors_from_database(self, *, caller: AuthToken, status: int = None,
                                    id_token: str = None) -> ResultActorAvro:
        result = ResultActorAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            act_list = None
            try:
                act_list = self.get_container_management_database().get_actors()
            except Exception as e:
                self.logger.error("get_actors_from_database {}".format(e))
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.interpret(exception=e))
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)
                return result

            if act_list is not None:
                if status is None:
                    result.actors = Converter.fill_actors_from_db(act_list=act_list)
                else:
                    result.actors = Converter.fill_actors_from_db_status(act_list=act_list, status=status)

        except Exception as e:
            self.logger.error("get_actors_from_database {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_actors_from_database(self, *, caller: AuthToken, id_token: str = None) -> ResultActorAvro:
        return self.do_get_actors_from_database(caller=caller, id_token=id_token)

    def get_actors_from_database_name_type_status(self, *, name: str, actor_type: int, status: int,
                                                  caller: AuthToken, id_token: str = None) -> ResultActorAvro:
        return self.do_get_actors_from_database(caller=caller, id_token=id_token, status=status)

    def get_controllers(self, *, caller: AuthToken, id_token: str = None) -> ResultActorAvro:
        return self.do_get_actors(atype=ActorType.Orchestrator.value, caller=caller, id_token=id_token)

    def get_brokers(self, *, caller: AuthToken, id_token: str = None) -> ResultActorAvro:
        return self.do_get_actors(atype=ActorType.Broker.value, caller=caller, id_token=id_token)

    def get_authorities(self, *, caller: AuthToken, id_token: str = None) -> ResultActorAvro:
        return self.do_get_actors(atype=ActorType.Authority.value, caller=caller, id_token=id_token)

    def get_management_object(self, *, key: ID) -> ABCManagementObject:
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        return GlobalsSingleton.get().get_container().get_management_object_manager().get_management_object(key=key)

    def do_get_proxies(self, *, atype: ActorType, protocol: str, caller: AuthToken,
                       id_token: str = None) -> ResultProxyAvro:
        result = ResultProxyAvro()
        result.status = ResultAvro()

        if caller is None or protocol is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            proxies = None
            if atype == ActorType.Broker:
                proxies = ActorRegistrySingleton.get().get_broker_proxies(protocol=protocol)
            elif atype == ActorType.Authority:
                proxies = ActorRegistrySingleton.get().get_site_proxies(protocol=protocol)
            else:
                proxies = ActorRegistrySingleton.get().get_proxies(protocol=protocol)

            result.proxies = Converter.fill_proxies(proxies=proxies)
        except Exception as e:
            self.logger.error("get_broker_proxies {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.interpret(exception=e))
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_broker_proxies(self, *, protocol: str, caller: AuthToken, id_token: str = None) -> ResultProxyAvro:
        return self.do_get_proxies(atype=ActorType.Broker, protocol=protocol, caller=caller, id_token=id_token)

    def get_site_proxies(self, *, protocol: str, caller: AuthToken, id_token: str = None) -> ResultProxyAvro:
        return self.do_get_proxies(atype=ActorType.Authority, protocol=protocol, caller=caller, id_token=id_token)

    def get_proxies_by_protocol(self, *, protocol: str, caller: AuthToken, id_token: str = None) -> ResultProxyAvro:
        return self.do_get_proxies(atype=ActorType.All, protocol=protocol, caller=caller, id_token=id_token)



# coding: utf-8

from __future__ import absolute_import

from flask import json
from six import BytesIO

from fabric_cf.orchestrator.swagger_server.models.success import Success  # noqa: E501
from fabric_cf.orchestrator.swagger_server.test import BaseTestCase


class TestSlicesController(BaseTestCase):
    """SlicesController integration test stubs"""

    def test_slices_create_post(self):
        """Test case for slices_create_post

        Create slice
        """
        body = 'body_example'
        query_string = [('slice_name', 'slice_name_example'),
                        ('ssh_key', 'ssh_key_example'),
                        ('lease_end_time', 'lease_end_time_example')]
        response = self.client.open(
            '//slices/create',
            method='POST',
            data=json.dumps(body),
            content_type='text/plain',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_delete_slice_iddelete(self):
        """Test case for slices_delete_slice_iddelete

        Delete slice.
        """
        response = self.client.open(
            '//slices/delete/{slice_id}'.format(slice_id='slice_id_example'),
            method='DELETE')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_get(self):
        """Test case for slices_get

        Retrieve a listing of user slices
        """
        query_string = [('state', 'All')]
        response = self.client.open(
            '//slices',
            method='GET',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_modify_slice_idput(self):
        """Test case for slices_modify_slice_idput

        Modify slice
        """
        body = 'body_example'
        response = self.client.open(
            '//slices/modify/{slice_id}'.format(slice_id='slice_id_example'),
            method='PUT',
            data=json.dumps(body),
            content_type='text/plain')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_redeem_slice_idpost(self):
        """Test case for slices_redeem_slice_idpost

        Redeem resources reserved via Create API
        """
        response = self.client.open(
            '//slices/redeem/{slice_id}'.format(slice_id='slice_id_example'),
            method='POST')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_renew_slice_idpost(self):
        """Test case for slices_renew_slice_idpost

        Renew slice
        """
        query_string = [('new_lease_end_time', 'new_lease_end_time_example')]
        response = self.client.open(
            '//slices/renew/{slice_id}'.format(slice_id='slice_id_example'),
            method='POST',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_slice_idget(self):
        """Test case for slices_slice_idget

        slice properties
        """
        response = self.client.open(
            '//slices/{slice_id}'.format(slice_id='slice_id_example'),
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_slices_status_slice_idget(self):
        """Test case for slices_status_slice_idget

        slice status
        """
        response = self.client.open(
            '//slices/status/{slice_id}'.format(slice_id='slice_id_example'),
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))


if __name__ == '__main__':
    import unittest
    unittest.main()
